@@ -3,31 +3,21 @@ from __future__ import annotations
 import asyncio
 import functools
 import logging
+import os
 import random
+import traceback
 import urllib.parse
 import warnings
+from collections.abc import AsyncIterator, Generator, Sequence
 from types import TracebackType
-from typing import (
-    Any,
-    AsyncIterator,
-    Callable,
-    Generator,
-    List,
-    Optional,
-    Sequence,
-    Tuple,
-    Type,
-    cast,
-)
+from typing import Any, Callable, cast
 
+from ..asyncio.compatibility import asyncio_timeout
 from ..datastructures import Headers, HeadersLike
 from ..exceptions import (
-    InvalidHandshake,
     InvalidHeader,
-    InvalidMessage,
-    InvalidStatusCode,
+    InvalidHeaderValue,
     NegotiationError,
-    RedirectHandshake,
     SecurityError,
 )
 from ..extensions import ClientExtensionFactory, Extension
@@ -35,14 +25,16 @@ from ..extensions.permessage_deflate import enable_client_permessage_deflate
 from ..headers import (
     build_authorization_basic,
     build_extension,
+    build_host,
     build_subprotocol,
     parse_extension,
     parse_subprotocol,
     validate_subprotocols,
 )
-from ..http import USER_AGENT, build_host
+from ..http11 import USER_AGENT
 from ..typing import ExtensionHeader, LoggerLike, Origin, Subprotocol
 from ..uri import WebSocketURI, parse_uri
+from .exceptions import InvalidMessage, InvalidStatusCode, RedirectHandshake
 from .handshake import build_request, check_response
 from .http import read_response
 from .protocol import WebSocketCommonProtocol
@@ -58,18 +50,19 @@ class WebSocketClientProtocol(WebSocketCommonProtocol):
     :class:`WebSocketClientProtocol` provides :meth:`recv` and :meth:`send`
     coroutines for receiving and sending messages.
 
-    It supports asynchronous iteration to receive incoming messages::
+    It supports asynchronous iteration to receive messages::
 
         async for message in websocket:
             await process(message)
 
     The iterator exits normally when the connection is closed with close code
-    1000 (OK) or 1001 (going away). It raises
+    1000 (OK) or 1001 (going away) or without a close code. It raises
     a :exc:`~websockets.exceptions.ConnectionClosedError` when the connection
     is closed with any other code.
 
     See :func:`connect` for the documentation of ``logger``, ``origin``,
-    ``extensions``, ``subprotocols``, and ``extra_headers``.
+    ``extensions``, ``subprotocols``, ``extra_headers``, and
+    ``user_agent_header``.
 
     See :class:`~websockets.legacy.protocol.WebSocketCommonProtocol` for the
     documentation of ``ping_interval``, ``ping_timeout``, ``close_timeout``,
@@ -83,11 +76,12 @@ class WebSocketClientProtocol(WebSocketCommonProtocol):
     def __init__(
         self,
         *,
-        logger: Optional[LoggerLike] = None,
-        origin: Optional[Origin] = None,
-        extensions: Optional[Sequence[ClientExtensionFactory]] = None,
-        subprotocols: Optional[Sequence[Subprotocol]] = None,
-        extra_headers: Optional[HeadersLike] = None,
+        logger: LoggerLike | None = None,
+        origin: Origin | None = None,
+        extensions: Sequence[ClientExtensionFactory] | None = None,
+        subprotocols: Sequence[Subprotocol] | None = None,
+        extra_headers: HeadersLike | None = None,
+        user_agent_header: str | None = USER_AGENT,
         **kwargs: Any,
     ) -> None:
         if logger is None:
@@ -97,6 +91,7 @@ class WebSocketClientProtocol(WebSocketCommonProtocol):
         self.available_extensions = extensions
         self.available_subprotocols = subprotocols
         self.extra_headers = extra_headers
+        self.user_agent_header = user_agent_header
 
     def write_http_request(self, path: str, headers: Headers) -> None:
         """
@@ -118,7 +113,7 @@ class WebSocketClientProtocol(WebSocketCommonProtocol):
 
         self.transport.write(request.encode())
 
-    async def read_http_response(self) -> Tuple[int, Headers]:
+    async def read_http_response(self) -> tuple[int, Headers]:
         """
         Read status line and headers from the HTTP response.
 
@@ -126,16 +121,12 @@ class WebSocketClientProtocol(WebSocketCommonProtocol):
         after this coroutine returns.
 
         Raises:
-            InvalidMessage: if the HTTP message is malformed or isn't an
+            InvalidMessage: If the HTTP message is malformed or isn't an
                 HTTP/1.1 GET response.
 
         """
         try:
             status_code, reason, headers = await read_response(self.reader)
-        # Remove this branch when dropping support for Python < 3.8
-        # because CancelledError no longer inherits Exception.
-        except asyncio.CancelledError:  # pragma: no cover
-            raise
         except Exception as exc:
             raise InvalidMessage("did not receive a valid HTTP response") from exc
 
@@ -151,8 +142,8 @@ class WebSocketClientProtocol(WebSocketCommonProtocol):
     @staticmethod
     def process_extensions(
         headers: Headers,
-        available_extensions: Optional[Sequence[ClientExtensionFactory]],
-    ) -> List[Extension]:
+        available_extensions: Sequence[ClientExtensionFactory] | None,
+    ) -> list[Extension]:
         """
         Handle the Sec-WebSocket-Extensions HTTP response header.
 
@@ -179,23 +170,20 @@ class WebSocketClientProtocol(WebSocketCommonProtocol):
         order of extensions, may be implemented by overriding this method.
 
         """
-        accepted_extensions: List[Extension] = []
+        accepted_extensions: list[Extension] = []
 
         header_values = headers.get_all("Sec-WebSocket-Extensions")
 
         if header_values:
-
             if available_extensions is None:
-                raise InvalidHandshake("no extensions supported")
+                raise NegotiationError("no extensions supported")
 
-            parsed_header_values: List[ExtensionHeader] = sum(
+            parsed_header_values: list[ExtensionHeader] = sum(
                 [parse_extension(header_value) for header_value in header_values], []
             )
 
             for name, response_params in parsed_header_values:
-
                 for extension_factory in available_extensions:
-
                     # Skip non-matching extensions based on their name.
                     if extension_factory.name != name:
                         continue
@@ -226,8 +214,8 @@ class WebSocketClientProtocol(WebSocketCommonProtocol):
 
     @staticmethod
     def process_subprotocol(
-        headers: Headers, available_subprotocols: Optional[Sequence[Subprotocol]]
-    ) -> Optional[Subprotocol]:
+        headers: Headers, available_subprotocols: Sequence[Subprotocol] | None
+    ) -> Subprotocol | None:
         """
         Handle the Sec-WebSocket-Protocol HTTP response header.
 
@@ -236,22 +224,23 @@ class WebSocketClientProtocol(WebSocketCommonProtocol):
         Return the selected subprotocol.
 
         """
-        subprotocol: Optional[Subprotocol] = None
+        subprotocol: Subprotocol | None = None
 
         header_values = headers.get_all("Sec-WebSocket-Protocol")
 
         if header_values:
-
             if available_subprotocols is None:
-                raise InvalidHandshake("no subprotocols supported")
+                raise NegotiationError("no subprotocols supported")
 
             parsed_header_values: Sequence[Subprotocol] = sum(
                 [parse_subprotocol(header_value) for header_value in header_values], []
             )
 
             if len(parsed_header_values) > 1:
-                subprotocols = ", ".join(parsed_header_values)
-                raise InvalidHandshake(f"multiple subprotocols: {subprotocols}")
+                raise InvalidHeaderValue(
+                    "Sec-WebSocket-Protocol",
+                    f"multiple values: {', '.join(parsed_header_values)}",
+                )
 
             subprotocol = parsed_header_values[0]
 
@@ -263,25 +252,25 @@ class WebSocketClientProtocol(WebSocketCommonProtocol):
     async def handshake(
         self,
         wsuri: WebSocketURI,
-        origin: Optional[Origin] = None,
-        available_extensions: Optional[Sequence[ClientExtensionFactory]] = None,
-        available_subprotocols: Optional[Sequence[Subprotocol]] = None,
-        extra_headers: Optional[HeadersLike] = None,
+        origin: Origin | None = None,
+        available_extensions: Sequence[ClientExtensionFactory] | None = None,
+        available_subprotocols: Sequence[Subprotocol] | None = None,
+        extra_headers: HeadersLike | None = None,
     ) -> None:
         """
         Perform the client side of the opening handshake.
 
         Args:
             wsuri: URI of the WebSocket server.
-            origin: value of the ``Origin`` header.
-            available_extensions: list of supported extensions, in order in
-                which they should be tried.
-            available_subprotocols: list of supported subprotocols, in order
-                of decreasing preference.
-            extra_headers: arbitrary HTTP headers to add to the request.
+            origin: Value of the ``Origin`` header.
+            extensions: List of supported extensions, in order in which they
+                should be negotiated and run.
+            subprotocols: List of supported subprotocols, in order of decreasing
+                preference.
+            extra_headers: Arbitrary HTTP headers to add to the handshake request.
 
         Raises:
-            InvalidHandshake: if the handshake fails.
+            InvalidHandshake: If the handshake fails.
 
         """
         request_headers = Headers()
@@ -314,7 +303,8 @@ class WebSocketClientProtocol(WebSocketCommonProtocol):
         if self.extra_headers is not None:
             request_headers.update(self.extra_headers)
 
-        request_headers.setdefault("User-Agent", USER_AGENT)
+        if self.user_agent_header:
+            request_headers.setdefault("User-Agent", self.user_agent_header)
 
         self.write_http_request(wsuri.resource_name, request_headers)
 
@@ -348,7 +338,7 @@ class Connect:
 
     :func:`connect` can be used as a asynchronous context manager::
 
-        async with websockets.connect(...) as websocket:
+        async with connect(...) as websocket:
             ...
 
     The connection is closed automatically when exiting the context.
@@ -356,7 +346,7 @@ class Connect:
     :func:`connect` can be used as an infinite asynchronous iterator to
     reconnect automatically on errors::
 
-        async for websocket in websockets.connect(...):
+        async for websocket in connect(...):
             try:
                 ...
             except websockets.ConnectionClosed:
@@ -375,25 +365,26 @@ class Connect:
 
     Args:
         uri: URI of the WebSocket server.
-        create_protocol: factory for the :class:`asyncio.Protocol` managing
-            the connection; defaults to :class:`WebSocketClientProtocol`; may
-            be set to a wrapper or a subclass to customize connection handling.
-        logger: logger for this connection;
-            defaults to ``logging.getLogger("websockets.client")``;
-            see the :doc:`logging guide <../topics/logging>` for details.
-        compression: shortcut that enables the "permessage-deflate" extension
-            by default; may be set to :obj:`None` to disable compression;
-            see the :doc:`compression guide <../topics/compression>` for details.
-        origin: value of the ``Origin`` header. This is useful when connecting
-            to a server that validates the ``Origin`` header to defend against
-            Cross-Site WebSocket Hijacking attacks.
-        extensions: list of supported extensions, in order in which they
-            should be tried.
-        subprotocols: list of supported subprotocols, in order of decreasing
+        create_protocol: Factory for the :class:`asyncio.Protocol` managing
+            the connection. It defaults to :class:`WebSocketClientProtocol`.
+            Set it to a wrapper or a subclass to customize connection handling.
+        logger: Logger for this client.
+            It defaults to ``logging.getLogger("websockets.client")``.
+            See the :doc:`logging guide <../../topics/logging>` for details.
+        compression: The "permessage-deflate" extension is enabled by default.
+            Set ``compression`` to :obj:`None` to disable it. See the
+            :doc:`compression guide <../../topics/compression>` for details.
+        origin: Value of the ``Origin`` header, for servers that require it.
+        extensions: List of supported extensions, in order in which they
+            should be negotiated and run.
+        subprotocols: List of supported subprotocols, in order of decreasing
             preference.
-        extra_headers: arbitrary HTTP headers to add to the request.
-        open_timeout: timeout for opening the connection in seconds;
-            :obj:`None` to disable the timeout
+        extra_headers: Arbitrary HTTP headers to add to the handshake request.
+        user_agent_header: Value of  the ``User-Agent`` request header.
+            It defaults to ``"Python/x.y.z websockets/X.Y"``.
+            Setting it to :obj:`None` removes the header.
+        open_timeout: Timeout for opening the connection in seconds.
+            :obj:`None` disables the timeout.
 
     See :class:`~websockets.legacy.protocol.WebSocketCommonProtocol` for the
     documentation of ``ping_interval``, ``ping_timeout``, ``close_timeout``,
@@ -414,41 +405,40 @@ class Connect:
       the TCP connection. The host name from ``uri`` is still used in the TLS
       handshake for secure connections and in the ``Host`` header.
 
-    Returns:
-        WebSocketClientProtocol: WebSocket connection.
-
     Raises:
-        InvalidURI: if ``uri`` isn't a valid WebSocket URI.
-        InvalidHandshake: if the opening handshake fails.
-        ~asyncio.TimeoutError: if the opening handshake times out.
+        InvalidURI: If ``uri`` isn't a valid WebSocket URI.
+        OSError: If the TCP connection fails.
+        InvalidHandshake: If the opening handshake fails.
+        ~asyncio.TimeoutError: If the opening handshake times out.
 
     """
 
-    MAX_REDIRECTS_ALLOWED = 10
+    MAX_REDIRECTS_ALLOWED = int(os.environ.get("WEBSOCKETS_MAX_REDIRECTS", "10"))
 
     def __init__(
         self,
         uri: str,
         *,
-        create_protocol: Optional[Callable[[Any], WebSocketClientProtocol]] = None,
-        logger: Optional[LoggerLike] = None,
-        compression: Optional[str] = "deflate",
-        origin: Optional[Origin] = None,
-        extensions: Optional[Sequence[ClientExtensionFactory]] = None,
-        subprotocols: Optional[Sequence[Subprotocol]] = None,
-        extra_headers: Optional[HeadersLike] = None,
-        open_timeout: Optional[float] = 10,
-        ping_interval: Optional[float] = 20,
-        ping_timeout: Optional[float] = 20,
-        close_timeout: Optional[float] = None,
-        max_size: Optional[int] = 2 ** 20,
-        max_queue: Optional[int] = 2 ** 5,
-        read_limit: int = 2 ** 16,
-        write_limit: int = 2 ** 16,
+        create_protocol: Callable[..., WebSocketClientProtocol] | None = None,
+        logger: LoggerLike | None = None,
+        compression: str | None = "deflate",
+        origin: Origin | None = None,
+        extensions: Sequence[ClientExtensionFactory] | None = None,
+        subprotocols: Sequence[Subprotocol] | None = None,
+        extra_headers: HeadersLike | None = None,
+        user_agent_header: str | None = USER_AGENT,
+        open_timeout: float | None = 10,
+        ping_interval: float | None = 20,
+        ping_timeout: float | None = 20,
+        close_timeout: float | None = None,
+        max_size: int | None = 2**20,
+        max_queue: int | None = 2**5,
+        read_limit: int = 2**16,
+        write_limit: int = 2**16,
         **kwargs: Any,
     ) -> None:
         # Backwards compatibility: close_timeout used to be called timeout.
-        timeout: Optional[float] = kwargs.pop("timeout", None)
+        timeout: float | None = kwargs.pop("timeout", None)
         if timeout is None:
             timeout = 10
         else:
@@ -458,7 +448,7 @@ class Connect:
             close_timeout = timeout
 
         # Backwards compatibility: create_protocol used to be called klass.
-        klass: Optional[Type[WebSocketClientProtocol]] = kwargs.pop("klass", None)
+        klass: type[WebSocketClientProtocol] | None = kwargs.pop("klass", None)
         if klass is None:
             klass = WebSocketClientProtocol
         else:
@@ -471,7 +461,7 @@ class Connect:
         legacy_recv: bool = kwargs.pop("legacy_recv", False)
 
         # Backwards compatibility: the loop parameter used to be supported.
-        _loop: Optional[asyncio.AbstractEventLoop] = kwargs.pop("loop", None)
+        _loop: asyncio.AbstractEventLoop | None = kwargs.pop("loop", None)
         if _loop is None:
             loop = asyncio.get_event_loop()
         else:
@@ -495,6 +485,9 @@ class Connect:
         if subprotocols is not None:
             validate_subprotocols(subprotocols)
 
+        # Help mypy and avoid this error: "type[WebSocketClientProtocol] |
+        # Callable[..., WebSocketClientProtocol]" not callable  [misc]
+        create_protocol = cast(Callable[..., WebSocketClientProtocol], create_protocol)
         factory = functools.partial(
             create_protocol,
             logger=logger,
@@ -502,6 +495,7 @@ class Connect:
             extensions=extensions,
             subprotocols=subprotocols,
             extra_headers=extra_headers,
+            user_agent_header=user_agent_header,
             ping_interval=ping_interval,
             ping_timeout=ping_timeout,
             close_timeout=close_timeout,
@@ -517,18 +511,20 @@ class Connect:
         )
 
         if kwargs.pop("unix", False):
-            path: Optional[str] = kwargs.pop("path", None)
+            path: str | None = kwargs.pop("path", None)
             create_connection = functools.partial(
                 loop.create_unix_connection, factory, path, **kwargs
             )
         else:
-            host: Optional[str]
-            port: Optional[int]
+            host: str | None
+            port: int | None
             if kwargs.get("sock") is None:
                 host, port = wsuri.host, wsuri.port
             else:
                 # If sock is given, host and port shouldn't be specified.
                 host, port = None, None
+                if kwargs.get("ssl"):
+                    kwargs.setdefault("server_hostname", wsuri.host)
             # If host and port are given, override values from the URI.
             host = kwargs.pop("host", host)
             port = kwargs.pop("port", port)
@@ -558,21 +554,27 @@ class Connect:
             raise SecurityError("redirect from WSS to WS")
 
         same_origin = (
-            old_wsuri.host == new_wsuri.host and old_wsuri.port == new_wsuri.port
+            old_wsuri.secure == new_wsuri.secure
+            and old_wsuri.host == new_wsuri.host
+            and old_wsuri.port == new_wsuri.port
         )
 
-        # Rewrite the host and port arguments for cross-origin redirects.
+        # Rewrite secure, host, and port for cross-origin redirects.
         # This preserves connection overrides with the host and port
         # arguments if the redirect points to the same host and port.
         if not same_origin:
-            # Replace the host and port argument passed to the protocol factory.
             factory = self._create_connection.args[0]
+            # Support TLS upgrade.
+            if not old_wsuri.secure and new_wsuri.secure:
+                factory.keywords["secure"] = True
+                self._create_connection.keywords.setdefault("ssl", True)
+            # Replace secure, host, and port arguments of the protocol factory.
             factory = functools.partial(
                 factory.func,
                 *factory.args,
                 **dict(factory.keywords, host=new_wsuri.host, port=new_wsuri.port),
             )
-            # Replace the host and port argument passed to create_connection.
+            # Replace secure, host, and port arguments of create_connection.
             self._create_connection = functools.partial(
                 self._create_connection.func,
                 *(factory, new_wsuri.host, new_wsuri.port),
@@ -585,37 +587,35 @@ class Connect:
 
     # async for ... in connect(...):
 
-    BACKOFF_MIN = 1.92
-    BACKOFF_MAX = 60.0
-    BACKOFF_FACTOR = 1.618
-    BACKOFF_INITIAL = 5
+    BACKOFF_INITIAL = float(os.environ.get("WEBSOCKETS_BACKOFF_INITIAL_DELAY", "5"))
+    BACKOFF_MIN = float(os.environ.get("WEBSOCKETS_BACKOFF_MIN_DELAY", "3.1"))
+    BACKOFF_MAX = float(os.environ.get("WEBSOCKETS_BACKOFF_MAX_DELAY", "90.0"))
+    BACKOFF_FACTOR = float(os.environ.get("WEBSOCKETS_BACKOFF_FACTOR", "1.618"))
 
     async def __aiter__(self) -> AsyncIterator[WebSocketClientProtocol]:
-        backoff_delay = self.BACKOFF_MIN
+        backoff_delay = self.BACKOFF_MIN / self.BACKOFF_FACTOR
         while True:
             try:
                 async with self as protocol:
                     yield protocol
-            # Remove this branch when dropping support for Python < 3.8
-            # because CancelledError no longer inherits Exception.
-            except asyncio.CancelledError:  # pragma: no cover
-                raise
-            except Exception:
+            except Exception as exc:
                 # Add a random initial delay between 0 and 5 seconds.
-                # See 7.2.3. Recovering from Abnormal Closure in RFC 6544.
+                # See 7.2.3. Recovering from Abnormal Closure in RFC 6455.
                 if backoff_delay == self.BACKOFF_MIN:
                     initial_delay = random.random() * self.BACKOFF_INITIAL
                     self.logger.info(
-                        "! connect failed; reconnecting in %.1f seconds",
+                        "connect failed; reconnecting in %.1f seconds: %s",
                         initial_delay,
-                        exc_info=True,
+                        # Remove first argument when dropping Python 3.9.
+                        traceback.format_exception_only(type(exc), exc)[0].strip(),
                     )
                     await asyncio.sleep(initial_delay)
                 else:
                     self.logger.info(
-                        "! connect failed again; retrying in %d seconds",
+                        "connect failed again; retrying in %d seconds: %s",
                         int(backoff_delay),
-                        exc_info=True,
+                        # Remove first argument when dropping Python 3.9.
+                        traceback.format_exception_only(type(exc), exc)[0].strip(),
                     )
                     await asyncio.sleep(int(backoff_delay))
                 # Increase delay with truncated exponential backoff.
@@ -633,9 +633,9 @@ class Connect:
 
     async def __aexit__(
         self,
-        exc_type: Optional[Type[BaseException]],
-        exc_value: Optional[BaseException],
-        traceback: Optional[TracebackType],
+        exc_type: type[BaseException] | None,
+        exc_value: BaseException | None,
+        traceback: TracebackType | None,
     ) -> None:
         await self.protocol.close()
 
@@ -643,19 +643,12 @@ class Connect:
 
     def __await__(self) -> Generator[Any, None, WebSocketClientProtocol]:
         # Create a suitable iterator by calling __await__ on a coroutine.
-        return self.__await_impl_timeout__().__await__()
-
-    async def __await_impl_timeout__(self) -> WebSocketClientProtocol:
-        return await asyncio.wait_for(self.__await_impl__(), self.open_timeout)
+        return self.__await_impl__().__await__()
 
     async def __await_impl__(self) -> WebSocketClientProtocol:
-        for redirects in range(self.MAX_REDIRECTS_ALLOWED):
-            transport, protocol = await self._create_connection()
-            # https://github.com/python/typeshed/pull/2756
-            transport = cast(asyncio.Transport, transport)
-            protocol = cast(WebSocketClientProtocol, protocol)
-
-            try:
+        async with asyncio_timeout(self.open_timeout):
+            for _redirects in range(self.MAX_REDIRECTS_ALLOWED):
+                _transport, protocol = await self._create_connection()
                 try:
                     await protocol.handshake(
                         self._wsuri,
@@ -664,17 +657,20 @@ class Connect:
                         available_subprotocols=protocol.available_subprotocols,
                         extra_headers=protocol.extra_headers,
                     )
-                except Exception:
+                except RedirectHandshake as exc:
+                    protocol.fail_connection()
+                    await protocol.wait_closed()
+                    self.handle_redirect(exc.uri)
+                # Avoid leaking a connected socket when the handshake fails.
+                except (Exception, asyncio.CancelledError):
                     protocol.fail_connection()
                     await protocol.wait_closed()
                     raise
                 else:
                     self.protocol = protocol
                     return protocol
-            except RedirectHandshake as exc:
-                self.handle_redirect(exc.uri)
-        else:
-            raise SecurityError("too many redirects")
+            else:
+                raise SecurityError("too many redirects")
 
     # ... = yield from connect(...) - remove when dropping Python < 3.10
 
@@ -685,7 +681,7 @@ connect = Connect
 
 
 def unix_connect(
-    path: Optional[str] = None,
+    path: str | None = None,
     uri: str = "ws://localhost/",
     **kwargs: Any,
 ) -> Connect:
@@ -700,7 +696,7 @@ def unix_connect(
     It's mainly useful for debugging servers listening on Unix sockets.
 
     Args:
-        path: file system path to the Unix socket.
+        path: File system path to the Unix socket.
         uri: URI of the WebSocket server; the host is used in the TLS
             handshake for secure connections and in the ``Host`` header.
 
